#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `terrascan` package."""

import pytest
from terrascan import terrascan


def test_with_empty_args():
    """
    User passes no args, should fail with SystemExit
    """
    with pytest.raises(SystemExit):
        terrascan.main(['-h'])

def test_success():
    """
    Test successful terraform templates
    """
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        terrascan.main([
            '-l',
            'tests/infrastructure/success',
        ])
    assert pytest_wrapped_e.value.code == 0
    
def test_fail():
    """
    Test successful terraform templates
    """
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        terrascan.main([
            '-l',
            'tests/infrastructure/fail',
        ])
    assert pytest_wrapped_e.value.code == 4


# -*- coding: utf-8 -*-
"""
    terrascan: A collection of security and best practice tests for static code analysis of terraform templates using terraform_validate.

    Copyright (C) 2017  Cesar Rodriguez

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import argparse
import unittest
import os
import re
import sys
import subprocess
import json
import time
from terrascan.embedded import terraform_validate
import logging

jsonOutput = {
    "dateTimeStamp": "",
    "terrascan-version": "",
    "failures": [],
    "errors": [],
    "files": [],
    "rules": []
}


###############################################################################################################################################################################
# Rules:  these are the rules used to verify that the terraform files are set up correctly.
#
# available methods for resources:
#   property(property_name): returns a list of the requested property_name; if self.v.error_if_property_missing() is called before the rule, will fail if property is missing.
#   with_property(property_name, regex_value): returns a list of the requested property_name with the requested regex_value.
#   should_not_exist(): fails if the resource doesn't exist.
#   should_have_properties(properties_list): fails if any of the properties in the given properties_list doesn't exist.
#   should_not_have_properties(properties_list): fails if any of the properties in the given properties_list exists.
#   find_property(property_name_regex): returns a list of the requested property_name_regex.
# available methods for properties:
#   property(property_name):  returns a list of the requested property_name; if self.v.error_if_property_missing() is called before the rule, will fail if property is missing.
#   should_equal(expected_value): fails if property value doesn't equal given expected_value.
#   should_equal_case_insensitive(expected_value): fails if property value doesn't equal given expected_value ignoring case.
#   should_not_equal(expected_value): fails if property value equals given expected_value.
#   should_not_equal_case_insensitive(expected_value): fails if property value equals given expected_value ignoring case.
#   list_should_contain_any(values_list): fails if the value of the property doesn't contain any of the values in values_list.
#   list_should_contain(values_list): fails if the value of the property doesn't contain all of the values in values_list.
#   list_should_not_contain(values_list): fails if the value of the property contains any of the values in values_list.
#   should_have_properties(properties_list): fails if a property doesn't contain any of the properties in properties_list.
#   should_not_have_properties(properties_list): fails if a property contains any of the properties in properties_list.
#   find_property(property_name_regex): returns a list of the requested property_name_regex.
#   should_match_regex(property_value_regex): fails if the value of the property doesn't match the given property_value_regex.
#   should_contain_valid_json(): fails if the value of the property doesn't contain valid json.
###############################################################################################################################################################################
class Rules(unittest.TestCase):
    
    rules = []

    def setUp(self):
        self.v = terraform_validate.Validator()
        self.v.preprocessor = self.preprocessor
        self.v.overrides = self.overrides

    #################################################################################################
    # examples of good and bad (marked with ***error***) are given before each rule
    #################################################################################################

    #################################################################################################
    # This resource block creates an S3 bucket with encryption.
    # resource "aws_s3_bucket" "encryptedBucket" {
    #   bucket = "good-bucket-name"
    #   server_side_encryption_configuration {
    #     rule {
    #       apply_server_side_encryption_by_default {
    #         kms_master_key_id = "${data.aws_kms_key.bucket.arn}"
    #         sse_algorithm = "aws:kms"
    #       }
    #     }
    #   }
    # }
    #
    # This resource block creates an S3 bucket with no encryption.  ***error***
    # resource "aws_s3_bucket" "noEncryption" {
    #   bucket = "bad-bucket-name"
    # }
    def test_aws_s3_bucket_server_side_encryption_configuration(self):
        # get name of rule from function name
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # verify that server side encryption is turned on for s3 buckets
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            # to change severity, override it here (default is high)
            validator.severity = "high"
            validator.resources('aws_s3_bucket').should_have_properties(['server_side_encryption_configuration'])

    #################################################################################################
    # This resource block creates a dynamodb table with no encryption.  ***error***
    # resource "aws_dynamodb_table" "noEncryption" {
    #   name = "${local.env}"
    # }
    #
    # This resource block creates a dynamodb table with no encryption.  ***error***
    # resource "aws_dynamodb_table" "encryptionEnabledFalse" {
    #   name = "${local.env}"
    #   server_side_encryption {
    #     enabled = false
    #   }
    # }
    #
    # This resource block creates a dynamodb table with encryption.
    # resource "aws_dynamodb_table" "encryptionEnabledTrue" {
    #   name = "${local.env}"
    #   server_side_encryption {
    #     enabled = true
    #   }
    # }
    def test_aws_dynamodb_table_encryption(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # verify that encryption is turned on for dynamodb tables
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_dynamodb_table').property('server_side_encryption').property('enabled').should_equal(True)

    #################################################################################################
    # This resource block creates an ebs volue with no encryption.  ***error***
    # resource "aws_ebs_volume" "noEncryption" {
    #   availability_zone = "us-east-1a"
    #   size = 10
    #   type = "gp2"
    #   tags {
    #     Name = "Encryption Test"
    #   }
    #   kms_key_id = "${data.aws_kms_key.volume.arn}"
    # }
    #
    # This resource block creates an ebs volue with no encryption.  ***error***
    # resource "aws_ebs_volume" "encryptionEnabledFalse" {
    #   availability_zone = "us-east-1a"
    #   size = 10
    #   type = "gp2"
    #   tags {
    #     Name = "Encryption Test"
    #   }
    #   encrypted = false
    #   kms_key_id = "${data.aws_kms_key.volume.arn}"
    # }
    #
    # resource "aws_ebs_volume" "encryptionEnabledTrue" {
    #   availability_zone = "us-east-1a"
    #   size = 10
    #   type = "gp2"
    #   tags {
    #     Name = "Encryption Test"
    #   }
    #   # The attributes below mark the volume for encryption.
    #   encrypted = true
    #   kms_key_id = "${data.aws_kms_key.volume.arn}"
    # }
    def test_aws_ebs_volume_encryption(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # verify that all resources of type 'aws_ebs_volume' are encrypted
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_ebs_volume').property('encrypted').should_equal(True)

    #################################################################################################
    # This resource block creates a kms key without key rotation.  ***error***
    # resource "aws_kms_key" "no_enable_key_rotation" {
    #   description             = "KMS key 1"
    #   deletion_window_in_days = 10
    # }
    #
    # This resource block creates a kms key without key rotation.  ***error***
    # resource "aws_kms_key" "enable_key_rotation_false" {
    #   description             = "KMS key 1"
    #   deletion_window_in_days = 10
    #   enable_key_rotation     = false
    # }
    #
    # This resource block creates a kms key with key rotation.
    # resource "aws_kms_key" "enable_key_rotation_true" {
    #   description             = "KMS key 1"
    #   deletion_window_in_days = 10
    #   enable_key_rotation     = true
    # }
    def test_aws_kms_key_rotation(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # verify that all aws_kms_key resources have key rotation enabled
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_kms_key').property('enable_key_rotation').should_equal(True)

    #################################################################################################
    # This resource block creates an iam user login profile.  ***error***
    # resource "aws_iam_user_login_profile" "badExample" {
    #   user    = "some_user_name"
    #   pgp_key = "keybase:some_person_that_exists"
    # }
    def test_aws_iam_user_login_profile(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # resource aws_iam_user_login_profile should not exist
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_iam_user_login_profile').should_not_exist()

    #################################################################################################
    # This resource block creates a security group rule.  ***error***
    # resource "aws_security_group_rule" "badExample" {
    #   type              = "ingress"
    #   from_port         = 0
    #   to_port           = 65535
    #   protocol          = "tcp"
    #   cidr_blocks       = ["0.0.0.0/0"]
    #   self              = true
    #   security_group_id = "${aws_security_group.emr-master.id}"
    # }
    def test_aws_security_group_rule_ingress_open(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # verify that security group rule ingress is not open to 0.0.0.0/0
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_security_group_rule').with_property('type', 'ingress').property('cidr_blocks').list_should_not_contain('0.0.0.0/0')

    #################################################################################################
    # This resource block creates a security group.  ***error***
    # resource "aws_security_group" "badExample" {
    #   name        = "generic-emr-master"
    #   description = "Manage traffic for EMR masters"
    #   vpc_id      = "${local.emr_vpc_id}"
    #   ingress {
    #     from_port = 443
    #     to_port = 443
    #     protocol = "tcp"
    #     cidr_blocks = ["0.0.0.0/0"]
    #   }
    # }
    def test_aws_security_group_ingress_open(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # verify that security group ingress is not open to 0.0.0.0/0
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_security_group').property('ingress').property('cidr_blocks').list_should_not_contain('0.0.0.0/0')

    #################################################################################################
    # This resource block creates an aws_db_instance.
    # resource "aws_db_instance" "default" {
    #   allocated_storage    = 20
    #   engine               = "mysql"
    #   instance_class       = "db.t2.micro"
    #   name                 = "mydb"
    #   storage_encrypted    = true
    # }
    #
    # This resource block creates an aws_db_instance.  ***error***
    # resource "aws_db_instance" "default" {
    #   allocated_storage    = 20
    #   engine               = "mysql"
    #   instance_class       = "db.t2.micro"
    #   name                 = "mydb"
    #   storage_encrypted    = false
    # }
    def test_aws_db_instance_encrypted(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert that DB is encrypted
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_db_instance').property('storage_encrypted').should_equal(True)

    #################################################################################################
    # This resource block creates an aws_rds_cluster.
    # resource "aws_rds_cluster" "default" {
    #   database_name           = "mydb"
    #   master_username         = "foo"
    #   master_password         = "bar"
    #   storage_encrypted       = true
    # }
    #
    # This resource block creates an aws_rds_cluster.  ***error***
    # resource "aws_rds_cluster" "default" {
    #   database_name           = "mydb"
    #   master_username         = "foo"
    #   master_password         = "bar"
    # }
    def test_aws_rds_cluster_encryption(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert resource is encrypted
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_rds_cluster').property('storage_encrypted').should_equal(True)

    #################################################################################################
    # public exposure - these were part of the original terrascan
    #################################################################################################

    def test_aws_alb_public(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources(['aws_lb', 'aws_alb']).property('internal').should_not_equal(False)

    def test_aws_db_instance_public(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_db_instance').property('publicly_accessible').should_not_equal(True)

    def test_aws_dms_replication_instance_public(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_dms_replication_instance').property('publicly_accessible').should_not_equal(True)

    def test_aws_elb_public(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_elb').property('internal').should_not_equal(False)

    def test_aws_instance_public(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_instance').property('associate_public_ip_address').should_not_equal(True)

    def test_aws_launch_configuration_public(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_launch_configuration').property('associate_public_ip_address').should_not_equal(True)

    def test_aws_rds_cluster_instance_public(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_rds_cluster_instance').property('publicly_accessible').should_not_equal(True)

    def test_aws_redshift_cluster_public(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_redshift_cluster').property('publicly_accessible').should_not_equal(True)

    def test_aws_s3_bucket_public(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_s3_bucket').property('acl').should_not_equal('public-read')
            validator.resources('aws_s3_bucket').property('acl').should_not_equal('public-read-write')
            validator.resources('aws_s3_bucket').property('acl').should_not_equal('authenticated-read')
            validator.resources('aws_s3_bucket').should_not_have_properties(['website'])

    #################################################################################################
    # other terrascan original rules - prefix test with X to disable
    #################################################################################################
    # encryption
    #################################################################################################

    def test_aws_alb_listener_port(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert that listener port is 443
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources(['aws_lb_listener', 'aws_alb_listener']).property('port').should_equal('443')

    def test_aws_alb_listener_protocol(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert that protocol is not http
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources(['aws_lb_listener', 'aws_alb_listener']).property('protocol').should_not_equal_case_insensitive('http')

    def test_aws_alb_listener_ssl_policy(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert that old ssl policies are not used
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources(['aws_lb_listener', 'aws_alb_listener']).property('ssl_policy').should_not_equal('ELBSecurityPolicy-2015-05')
            validator.resources(['aws_lb_listener', 'aws_alb_listener']).property('ssl_policy').should_not_equal('ELBSecurityPolicy-TLS-1-0-2015-04')

    def test_aws_alb_listener_certificate(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert that certificate_arn is set
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources(['aws_lb_listener', 'aws_alb_listener']).should_have_properties(['certificate_arn'])

    def test_aws_ami_ebs_block_device_encryption(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert ami 'ebs_block_device' blocks are encrypted
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_ami').property('ebs_block_device').property('encrypted').should_equal(True)

    def test_aws_ami_ebs_block_device_kms(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert ami 'ebs_block_device' blocks has KMS
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_ami').property('ebs_block_device').should_have_properties(['kms_key_id'])

    def test_aws_ami_copy_encryption(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert resources are encrypted
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_ami_copy').property('encrypted').should_equal(True)

    def test_aws_ami_copy_kms(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert that a KMS key has been provided
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_ami_copy').should_have_properties(['kms_key_id'])

    def test_aws_api_gateway_domain_name_certificate(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert that certificate settings have been configured
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_api_gateway_domain_name').should_have_properties(
                [
                    'certificate_name',
                    'certificate_body',
                    'certificate_chain',
                    'certificate_private_key'
                ])

    def test_aws_instance_ebs_block_device_encrypted(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert ec2 instance 'ebs_block_device' is encrypted
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_instance').property('ebs_block_device').property('encrypted').should_equal(True)

    def test_aws_cloudfront_distribution_origin_protocol_policy(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert that origin receives https only traffic
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_cloudfront_distribution').property('origin').property('custom_origin_config').property('origin_protocol_policy').should_equal("https-only")

    def test_aws_cloudfront_distribution_def_cache_viewer_prot_policy(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert that cache protocol doesn't allow all
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_cloudfront_distribution').property('default_cache_behavior').property('viewer_protocol_policy').should_not_equal("allow-all")

    def test_aws_cloudfront_distribution_cache_beh_viewer_proto_policy(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert that cache protocol doesn't allow all
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_cloudfront_distribution').property('cache_behavior').property('viewer_protocol_policy').should_not_equal("allow-all")

    def test_aws_cloudtrail_kms(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert that a KMS key has been provided
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_cloudtrail').should_have_properties(['kms_key_id'])

    def test_aws_codebuild_project_kms(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert that a KMS key has been provided
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_codebuild_project').should_have_properties(['encryption_key'])

    def test_aws_codepipeline_kms(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert that a KMS key has been provided
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_codepipeline').should_have_properties(['encryption_key'])

    def test_aws_db_instance_kms(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert that a KMS key has been provided
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_db_instance').should_have_properties(['kms_key_id'])

    def test_aws_dms_endpoint_ssl_mode(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert that SSL is verified
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_dms_endpoint').property('ssl_mode').should_equal('verify-full')

    def test_aws_dms_endpoint_kms(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert that a KMS key has been provided
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_dms_endpoint').should_have_properties(
                [
                    'kms_key_arn'
                ])

    def test_aws_dms_endpoint_certificate(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert that SSL cert has been provided
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_dms_endpoint').should_have_properties(
                [
                    'certificate_arn'
                ])

    def test_aws_dms_replication_instance_kms(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert that a KMS key has been provided
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_dms_replication_instance').should_have_properties(['kms_key_arn'])

    def test_aws_ebs_volume_kms(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert that a KMS key has been provided
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_ebs_volume').should_have_properties(['kms_key_id'])

    def test_aws_efs_file_system_encryption(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert that all resources of type 'aws_efs_file_system' are encrypted
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_efs_file_system').property('encrypted').should_equal(True)

    def test_aws_efs_file_system_kms(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert that a KMS key has been provided
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_efs_file_system').should_have_properties(['kms_key_id'])

    def test_aws_elastictranscoder_pipeline_kms(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert that a KMS key has been provided
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_elastictranscoder_pipeline').should_have_properties(['aws_kms_key_arn'])

    def test_aws_elb_listener_port_80(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert ELB listener port is not 80 (http)
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_elb').property('listener').property('lb_port').should_not_equal(80)

    def test_aws_elb_listener_port_21(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert ELB listener port is not 21 ftp
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_elb').property('listener').property('lb_port').should_not_equal(21)

    def test_aws_elb_listener_port_23(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert ELB listener port is not 23 telnet
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_elb').property('listener').property('lb_port').should_not_equal(23)

    def test_aws_elb_listener_port_5900(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert ELB listener port is not 5900 VNC
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_elb').property('listener').property('lb_port').should_not_equal(5900)

    def test_aws_kinesis_firehose_delivery_stream_s3_kms(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert ELB listener port is not 80 (http)
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_kinesis_firehose_delivery_stream').property('s3_configuration').should_have_properties(['kms_key_arn'])

    def test_aws_kinesis_firehose_delivery_stream_extended_s3_kms(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert ELB listener port is not 80 (http)
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_kinesis_firehose_delivery_stream').property('extended_s3_configuration').should_have_properties(['kms_key_arn'])

    def test_aws_lambda_function_kms(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert that a KMS key has been provided
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_lambda_function').should_have_properties(['kms_key_arn'])

    def test_aws_opsworks_application_encryption(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert resource is encrypted
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_opsworks_application').property('enable_ssl').should_equal(True)

    def test_aws_rds_cluster_kms(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert resource has a KMS with CMKs
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_rds_cluster').should_have_properties(['kms_key_id'])

    def test_aws_redshift_cluster_encryption(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert resource is encrypted
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_redshift_cluster').property('encrypted').should_equal(True)

    def test_aws_redshift_cluster_kms(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert resource has a KMS with CMKs
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_redshift_cluster').should_have_properties(['kms_key_id'])

    def test_aws_s3_bucket_object_kms(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert resource has a KMS with CMKs
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_s3_bucket_object').should_have_properties(['kms_key_id'])

    def test_aws_sqs_queue_kms(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert resource has a KMS with CMK
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_sqs_queue').should_have_properties(['kms_master_key_id', 'kms_data_key_reuse_period_seconds'])

    def test_aws_ssm_parameter_encryption(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert resource is encrypted with KMS
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_ssm_parameter').property('type').should_equal("SecureString")

    def test_aws_ssm_parameter_kms(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # Assert resource has a KMS with CMK
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_ssm_parameter').should_have_properties(['key_id'])

    #################################################################################################
    # logging and monitoring
    #################################################################################################

    def test_aws_alb_logging(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources(['aws_lb', 'aws_alb']).should_have_properties(['access_logs'])

    def test_aws_cloudfront_distribution_logging(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_cloudfront_distribution').should_have_properties(['logging_config'])

    def test_aws_cloudtrail_logging(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_cloudtrail').property('enable_logging').should_not_equal(False)

    def test_aws_elb_logging(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_elb').should_have_properties(['access_logs'])

    def test_aws_emr_cluster_logging(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_emr_cluster').should_have_properties(['log_uri'])

    def test_aws_kinesis_firehose_delivery_stream__s3_config_logging(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_kinesis_firehose_delivery_stream').property('s3_configuration').should_have_properties(['cloudwatch_logging_options'])
            validator.resources('aws_kinesis_firehose_delivery_stream').property('s3_configuration').property('cloudwatch_logging_options').property('enabled').should_equal(True)

    def test_aws_kinesis_firehose_delivery_stream_redshift_conf_logging(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_kinesis_firehose_delivery_stream').property('redshift_configuration').should_have_properties(['cloudwatch_logging_options'])
            validator.resources('aws_kinesis_firehose_delivery_stream').property('redshift_configuration').property('cloudwatch_logging_options').property('enabled').should_equal(True)

    def test_aws_kinesis_firehose_delivery_stream__es_config_logging(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_kinesis_firehose_delivery_stream').property('elasticsearch_configuration').should_have_properties(['cloudwatch_logging_options'])
            validator.resources('aws_kinesis_firehose_delivery_stream').property('elasticsearch_configuration').property('cloudwatch_logging_options').property('enabled').should_equal(True)

    def test_aws_redshift_cluster_logging(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        self.v.error_if_property_missing()
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_redshift_cluster').property('enable_logging').should_not_equal(False)

    def test_aws_s3_bucket_logging(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_s3_bucket').should_have_properties(['logging'])

    def test_aws_ssm_maintenance_window_task_logging(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_ssm_maintenance_window_task').should_have_properties(['logging_info'])

    #################################################################################################
    # security group
    #################################################################################################

    def test_aws_db_security_group_used(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # This security group type exists outside of VPC (e.g. ec2 classic)
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_db_security_group').should_not_exist()

    def test_aws_redshift_security_group_used(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # This security group type exists outside of VPC (e.g. ec2 classic)
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_redshift_security_group').should_not_exist()

    def test_aws_elasticache_security_group_used(self):
        ruleName = sys._getframe().f_code.co_name[5:]
        self.rules.append(ruleName)
        # This security group type exists outside of VPC (e.g. ec2 classic)
        validator_generator = self.v.get_terraform_files(self.isRuleOverridden(ruleName))
        for validator in validator_generator:
            validator.resources('aws_elasticache_security_group').should_not_exist()


    def isRuleOverridden(self, ruleName):
        for override in self.overrides:
            if ruleName == override[0]:
                return True
        return False


#################################################################################################
# run the tests
#################################################################################################
def terrascan(args):
    start = time.time()

    try:
        result = subprocess.run(['pip', 'show', 'terrascan-sf'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout = result.stdout.decode("utf-8")
        versionStr = "Version: "
        startIndex = stdout.find(versionStr)
    except:
        startIndex = -1
    if startIndex == -1:
        version = "?"
    else:
        startIndex += len(versionStr)
        endIndex = stdout.find("\r", startIndex)
        version = stdout[startIndex:endIndex]
    
    # process the arguments
    if args.warranty or args.gpl:
        print("terrascan  Copyright (C) 2020  Cesar Rodriguez\n")
        if args.warranty:
            print("This program is distributed in the hope that it will be useful,")
            print("but WITHOUT ANY WARRANTY; without even the implied warranty of")
            print("MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the")
            print("GNU General Public License for more details.")
        else:
            print("This program is free software: you can redistribute it and/or modify")
            print("it under the terms of the GNU General Public License as published by")
            print("the Free Software Foundation, either version 3 of the License, or")
            print("(at your option) any later version.  see <http://www.gnu.org/licenses/>")

        sys.exit(0)

    terraformLocation = args.location[0]
    if not os.path.isabs(terraformLocation):
        terraformLocation = os.path.join(os.sep, os.path.abspath("."),  terraformLocation)
    if args.vars:
        variablesJsonFilename = []
        for fileName in args.vars:
            if not os.path.isabs(fileName):
                fileName = os.path.join(os.sep, os.path.abspath("."),  fileName)
            variablesJsonFilename.append(fileName)
    else:
        variablesJsonFilename = None
    if args.overrides:
        Rules.overrides = []
        overridesFileName = args.overrides[0]
        if not os.path.isabs(overridesFileName):
            overridesFileName = os.path.join(os.sep, os.path.abspath("."),  overridesFileName)
        try:
            with open(overridesFileName, "r", encoding="utf-8") as fp:
                overridesFileString = fp.read()
            overrides = json.loads(overridesFileString)
            overrides = overrides["overrides"]
            # validate overrides
            for override in overrides:
                if len(override) < 2 or len(override) > 3 or len(override) == 3 and not re.match(r"RR-\d{1,10}$|RAR-\d{1,10}$", override[2]):
                    print("***Invalid entry in overrides file:  " + override)
                    print("Needs to be in the following format:  rule_name:resource_name or rule_name:resource_name:RR-xxx or rule_name:resource_name:RAR-xxx where xxx is 1-10 digits")
                    sys.exit(99)
                Rules.overrides.append(override)
        except Exception as e:
            print("***Error loading overrides file " + overridesFileName)
            print(e)
            sys.exit(99)
    else:
        Rules.overrides = ""
    if args.results:
        outputJsonFileName = args.results[0]
        if not os.path.isabs(outputJsonFileName):
            outputJsonFileName = os.path.join(os.sep, os.path.abspath("."),  outputJsonFileName)
    else:
        outputJsonFileName = None
    if args.config:
        config = args.config[0]
    else:
        config = None
        
    # set logging based on logging.config if present (default is error)
    if config == "none":
        logging.basicConfig(level=logging.CRITICAL)
    elif config == "warning":
        logging.basicConfig(level=logging.WARNING)
    elif config == "info":
        logging.basicConfig(level=logging.INFO)
    elif config == "debug":
        logging.basicConfig(level=logging.DEBUG)
    else:
        config = "error"
        logging.basicConfig(level=logging.ERROR)

    print("terrascan version {0}".format(version))
    print("Logging level set to {0}.".format(config))

    Rules.preprocessor = terraform_validate.PreProcessor(jsonOutput)
    Rules.preprocessor.process(terraformLocation, variablesJsonFilename)

    runner = unittest.TextTestRunner()
    itersuite = unittest.TestLoader().loadTestsFromTestCase(Rules)
    runner.run(itersuite)

    end = time.time()
    elapsedTime = end - start
    processedMessage = "Processed on " + time.strftime("%m/%d/%Y") + " at " + time.strftime("%H:%M")

    if outputJsonFileName:
        jsonOutput["dateTimeStamp"] = processedMessage
        jsonOutput["terrascan-version"] = version
        for fileName in Rules.preprocessor.fileNames:
            jsonOutput["files"].append(fileName)
        for rule in Rules.rules:
            jsonOutput["rules"].append(rule)
        with open(outputJsonFileName, 'w') as jsonOutFile:
            json.dump(jsonOutput, jsonOutFile)

    print("\nProcessed " + str(len(Rules.preprocessor.fileNames)) + " files in " + terraformLocation + "\n")
    for fileName in Rules.preprocessor.fileNames:
        logging.debug("  Processed " + fileName)
    print("")

    print(processedMessage)
    print("Results (took %.2f seconds):" % elapsedTime)
    rc = 0
    print("\nFailures: (" + str(len(jsonOutput["failures"])) + ")")
    for failure in jsonOutput["failures"]:
        m, f = getMF(failure)
        waived = ""
        if len(failure["waived"]) > 0:
            waived =  failure["waived"] + " "
        print("[" + failure["severity"] + "] " + waived + failure["message"] + m + f)
        if failure["waived"] == "":
            rc = 4
    print("\nErrors: (" + str(len(jsonOutput["errors"])) + ")")
    for error in jsonOutput["errors"]:
        m, f = getMF(error)
        print("[" + error["severity"] + "] " + error["message"] + m + f)
        rc = 4
    if args.displayRules:
        print("\nRules used:")
        for rule in Rules.rules:
            print(rule)

    sys.exit(rc)


# Returns command line parser for terrascan
def create_parser():
    parser = argparse.ArgumentParser(description="A collection of security and best practice tests for static code analysis of terraform templates using terraform_validate.")

    # only required if optional parameters not present
    req = '-w' not in sys.argv and '--warranty' not in sys.argv and '-g' not in sys.argv and '--gpl' not in sys.argv

    parser.add_argument(
        '-l',
        '--location',
        help='location of terraform templates to scan',
        nargs=1,
        required=req
    )
    parser.add_argument(
        '-v',
        '--vars',
        help='variables json or .tf file name',
        nargs='*',
    )
    parser.add_argument(
        '-o',
        '--overrides',
        help='override rules file name',
        nargs=1
    )
    parser.add_argument(
        '-r',
        '--results',
        help='output results file name',
        nargs=1,    )
    parser.add_argument(
        '-d',
        '--displayRules',
        help='display the rules used',
        nargs='?',
        const=True, default=False
    )
    parser.add_argument(
        '-w',
        '--warranty',
        help='displays the warranty',
        nargs='?',
        const=True, default=False
    )
    parser.add_argument(
        '-g',
        '--gpl',
        help='displays license information',
        nargs='?',
        const=True, default=False
    )
    parser.add_argument(
        '-c',
        '--config',
        help='logging configuration:  error, warning, info, debug, or none; default is error',
        nargs=1,    )
    parser.set_defaults(func=terrascan)

    return parser


def getMF(json):
    if json["moduleName"] == "---":
        m = ""
    else:
        m = " in module " + json["moduleName"]
    if json["fileName"] == "---":
        f = ""
    else:
        if m == "":
            f = " in file " + json["fileName"]
        else:
            f = ", file " + json["fileName"]
    return m, f

def main(args=None):
    """
    Terrascan console script. Parses user input to determine location of
    terraform templates and which tests to execute
    """
    parser = create_parser()
    args = parser.parse_args(args)
    
    #tests = args.tests[0]
    #location = path.abspath(args.location[0])
    
    #if not path.exists(location):
        #print("ERROR: The specified location doesn't exists")
        #exit(1)
        
    #exit(run_test(location, tests))
    terrascan(args)



import hcl
import os
import re
import traceback
import json
import logging
import sys

class TerraformPropertyList:

    def __init__(self, validator):
        self.properties = []
        self.validator = validator

    def tfproperties(self):
        return self.properties

    def property(self, property_name):
        propList = TerraformPropertyList(self.validator)
        for prop in self.properties:
            pvList = []
            if type(prop.property_value) is list:
                pvList = prop.property_value
            else:
                pvList.append(prop.property_value)

            wasFound = False
            for pv in pvList:
                if type(pv) is dict and property_name in pv.keys():
                    wasFound = True
                    propList.properties.append(TerraformProperty(prop.resource_type,
                                               "{0}.{1}".format(prop.resource_name, prop.property_name),
                                               property_name,
                                               pv[property_name],
                                               prop.moduleName,
                                               prop.fileName))
                    
            if not wasFound and self.validator.raise_error_if_property_missing:                    
                self.validator.preprocessor.add_failure(
                    "[{0}.{1}] should have property: '{2}'".format(prop.resource_type, "{0}.{1}".format(prop.resource_name, prop.property_name), property_name),
                    prop.moduleName,
                    prop.fileName,
                    self.validator.severity,
                    self.validator.isRuleOverridden,
                    self.validator.overrides,
                    prop.resource_name)

        return propList

    def should_equal_case_insensitive(self, expected_value):
        self.should_equal(expected_value, True)

    def should_equal(self, expected_value, caseInsensitive=False):
        for prop in self.properties:

            expected_value = self.int2str(expected_value)
            prop.property_value = self.int2str(prop.property_value)
            expected_value = self.bool2str(expected_value)
            prop.property_value = self.bool2str(prop.property_value)

            if caseInsensitive:
                # make both actual and expected lower case so case won't matter
                pv = prop.property_value.lower()
                ev = expected_value.lower()
            else:
                pv = prop.property_value
                ev = expected_value

            if pv != ev:
                self.validator.preprocessor.add_failure("[{0}.{1}.{2}] should be '{3}'. Is: '{4}'".format(prop.resource_type,
                                                                                                          prop.resource_name,
                                                                                                          prop.property_name,
                                                                                                          expected_value,
                                                                                                          prop.property_value),
                                                        prop.moduleName,
                                                        prop.fileName,
                                                        self.validator.severity,
                                                        self.validator.isRuleOverridden,
                                                        self.validator.overrides,
                                                        prop.resource_name)

    def should_not_equal_case_insensitive(self, expected_value):
        self.should_not_equal(expected_value, True)

    def should_not_equal(self, expected_value, caseInsensitive=False):
        for prop in self.properties:

            prop.property_value = self.int2str(prop.property_value)
            expected_value = self.int2str(expected_value)
            expected_value = self.bool2str(expected_value)
            prop.property_value = self.bool2str(prop.property_value)

            if caseInsensitive:
                # make both actual and expected lower case so case won't matter
                pv = prop.property_value.lower()
                ev = expected_value.lower()
            else:
                pv = prop.property_value
                ev = expected_value

            if pv == ev:
                self.validator.preprocessor.add_failure("[{0}.{1}.{2}] should not be '{3}'. Is: '{4}'".format(prop.resource_type,
                                                                                                              prop.resource_name,
                                                                                                              prop.property_name,
                                                                                                              expected_value,
                                                                                                              prop.property_value),
                                                        prop.moduleName,
                                                        prop.fileName,
                                                        self.validator.severity,
                                                        self.validator.isRuleOverridden,
                                                        self.validator.overrides,
                                                        prop.resource_name)

    def list_should_contain_any(self, values_list):
        if type(values_list) is not list:
            values_list = [values_list]

        for prop in self.properties:
            property_value = prop.property_value
            if type(property_value) is not list:
                property_value = [property_value]
            for pv in property_value:
                if pv not in values_list:
                    if type(prop.property_value) is list:
                        prop.property_value = [str(x) for x in prop.property_value]  # fix 2.6/7
                    self.validator.preprocessor.add_failure("[{0}.{1}.{2}] '{3}' should have been one of '{4}'.".format(prop.resource_type,
                                                                                                                        prop.resource_name,
                                                                                                                        prop.property_name,
                                                                                                                        prop.property_value,
                                                                                                                        values_list),
                                                            prop.moduleName,
                                                            prop.fileName,
                                                            self.validator.severity,
                                                            self.validator.isRuleOverridden,
                                                            self.validator.overrides,
                                                            prop.resource_name)
                    break;

    def list_should_contain(self, values_list):
        if type(values_list) is not list:
            values_list = [values_list]

        for prop in self.properties:

            values_missing = []
            for value in values_list:
                if value not in prop.property_value:
                    values_missing.append(value)

            if len(values_missing) != 0:
                if type(prop.property_value) is list:
                    prop.property_value = [str(x) for x in prop.property_value]  # fix 2.6/7
                self.validator.preprocessor.add_failure("[{0}.{1}.{2}] '{3}' should contain '{4}'.".format(prop.resource_type,
                                                                                                           prop.resource_name,
                                                                                                           prop.property_name,
                                                                                                           prop.property_value,
                                                                                                           values_missing),
                                                        prop.moduleName,
                                                        prop.fileName,
                                                        self.validator.severity,
                                                        self.validator.isRuleOverridden,
                                                        self.validator.overrides,
                                                        prop.resource_name)

    def list_should_not_contain(self, values_list):
        if type(values_list) is not list:
            values_list = [values_list]

        for prop in self.properties:

            values_missing = []
            for value in values_list:
                if value in prop.property_value:
                    values_missing.append(value)

            if len(values_missing) != 0:
                if type(prop.property_value) is list:
                    prop.property_value = [str(x) for x in prop.property_value]  # fix 2.6/7
                self.validator.preprocessor.add_failure("[{0}.{1}.{2}] '{3}' should not contain '{4}'.".format(prop.resource_type,
                                                                                                               prop.resource_name,
                                                                                                               prop.property_name,
                                                                                                               prop.property_value,
                                                                                                               values_missing),
                                                        prop.moduleName,
                                                        prop.fileName,
                                                        self.validator.severity,
                                                        self.validator.isRuleOverridden,
                                                        self.validator.overrides,
                                                        prop.resource_name)

    def should_have_properties(self, properties_list):
        if type(properties_list) is not list:
            properties_list = [properties_list]

        for prop in self.properties:
            property_names = prop.property_value.keys()
            for required_property_name in properties_list:
                if required_property_name not in property_names:
                    self.validator.preprocessor.add_failure("[{0}.{1}.{2}] should have property: '{3}'".format(prop.resource_type,
                                                                                                               prop.resource_name,
                                                                                                               prop.property_name,
                                                                                                               required_property_name),
                                                            prop.moduleName,
                                                            prop.fileName,
                                                            self.validator.severity,
                                                            self.validator.isRuleOverridden,
                                                            self.validator.overrides,
                                                            prop.resource_name)

    def should_not_have_properties(self, properties_list):
        if type(properties_list) is not list:
            properties_list = [properties_list]

        for prop in self.properties:
            property_names = prop.property_value.keys()
            for excluded_property_name in properties_list:
                if excluded_property_name in property_names:
                    self.validator.preprocessor.add_failure("[{0}.{1}.{2}] should not have property: '{3}'".format(prop.resource_type,
                                                                                                                   prop.resource_name,
                                                                                                                   prop.property_name,
                                                                                                                   excluded_property_name),
                                                            prop.moduleName,
                                                            prop.fileName,
                                                            self.validator.severity,
                                                            self.validator.isRuleOverridden,
                                                            self.validator.overrides,
                                                            prop.resource_name)

    def find_property(self, regex):
        lst = TerraformPropertyList(self.validator)
        for prop in self.properties:
            for nested_property in prop.property_value:
                if self.validator.matches_regex_pattern(nested_property, regex):
                    lst.properties.append(TerraformProperty(prop.resource_type,
                                           "{0}.{1}".format(prop.resource_name, prop.property_name),
                                           nested_property,
                                           prop.property_value[nested_property],
                                           prop.moduleName,
                                           prop.fileName))
        return lst

    def should_match_regex(self, regex):
        for prop in self.properties:
            if not self.validator.matches_regex_pattern(prop.property_value, regex):
                self.validator.preprocessor.add_failure("[{0}.{1}] should match regex '{2}'".format(prop.resource_type, "{0}.{1}".format(prop.resource_name, prop.property_name), regex),
                                                        prop.moduleName,
                                                        prop.fileName,
                                                        self.validator.severity,
                                                        self.validator.isRuleOverridden,
                                                        self.validator.overrides,
                                                        prop.resource_name)

    def should_contain_valid_json(self):
        for prop in self.properties:
            try:
                json.loads(prop.property_value)
            except:
                self.validator.preprocessor.add_failure("[{0}.{1}.{2}] is not valid json".format(prop.resource_type, prop.resource_name, prop.property_name),
                                                        prop.moduleName,
                                                        prop.fileName,
                                                        self.validator.severity,
                                                        self.validator.isRuleOverridden,
                                                        self.validator.overrides,
                                                        prop.resource_name)

    def bool2str(self, b):
        if str(b).lower() in ["true"]:
            return "True"
        if str(b).lower() in ["false"]:
            return "False"
        return b

    def int2str(self, property_value):
        if type(property_value) is int:
            property_value = str(property_value)
        return property_value


class TerraformProperty:

    def __init__(self, resource_type, resource_name, property_name, property_value, moduleName, fileName):
        self.resource_type = resource_type
        self.resource_name = resource_name
        self.property_name = property_name
        self.property_value = property_value
        self.moduleName = moduleName
        self.fileName = fileName


class TerraformResource:

    def __init__(self, typ, name, config, fileName, moduleName):
        self.type = typ
        self.name = name
        self.config = config
        self.fileName = fileName
        self.moduleName = moduleName


class TerraformResourceList:

    def __init__(self, validator, requestedResourceType, resourceTypes, resources):
        self.validator = validator
        self.resource_list = []
        self.requestedResourceType = requestedResourceType

        resourcesByType = {}
        for resourceName in resources:
            resource = resources[resourceName]
            resourceType = resource.type
            resourcesByType[resourceType] = resourcesByType.get(resourceType, {})
            resourcesByType[resourceType][resourceName] = resource.config

        if type(requestedResourceType) is str:
            resourceTypes = []
            for resourceType in resourcesByType:
                if validator.matches_regex_pattern(resourceType, requestedResourceType):
                    resourceTypes.append(resourceType)
        elif requestedResourceType is not None:
            resourceTypes = requestedResourceType

        for resourceType in resourceTypes:
            if resourceType in resourcesByType.keys():
                for resourceName in resourcesByType[resourceType]:
                    self.resource_list.append(
                        TerraformResource(resourceType, resourceName, resourcesByType[resourceType][resourceName], resources[resourceName].fileName, resources[resourceName].moduleName))

        self.resource_types = resourceTypes

    def property(self, property_name):
        lst = TerraformPropertyList(self.validator)
        if len(self.resource_list) > 0:
            for resource in self.resource_list:
                if property_name in resource.config.keys():
                    lst.properties.append(TerraformProperty(resource.type, resource.name, property_name, resource.config[property_name], resource.moduleName, resource.fileName))
                elif self.validator.raise_error_if_property_missing:
                    self.validator.preprocessor.add_failure("[{0}.{1}] should have property: '{2}'".format(resource.type, resource.name, property_name),
                                                            resource.moduleName,
                                                            resource.fileName,
                                                            self.validator.severity,
                                                            self.validator.isRuleOverridden,
                                                            self.validator.overrides,
                                                            resource.name)

        return lst

    def find_property(self, regex):
        lst = TerraformPropertyList(self.validator)
        if len(self.resource_list) > 0:
            for resource in self.resource_list:
                for prop in resource.config:
                    if self.validator.matches_regex_pattern(prop, regex):
                        lst.properties.append(TerraformProperty(resource.type,
                                                                resource.name,
                                                                prop,
                                                                resource.config[prop],
                                                                resource.moduleName,
                                                                resource.fileName))
        return lst

    def with_property(self, property_name, regex):
        lst = TerraformResourceList(self.validator, None, self.resource_types, {})

        if len(self.resource_list) > 0:
            for resource in self.resource_list:
                for prop in resource.config:
                    if prop == property_name:
                        tf_property = TerraformProperty(resource.type, resource.name, property_name, resource.config[property_name], resource.moduleName, resource.fileName)
                        if self.validator.matches_regex_pattern(tf_property.property_value, regex):
                            lst.resource_list.append(resource)

        return lst

    def should_not_exist(self):
        for terraformResource in self.resource_list:
            if terraformResource.type == self.requestedResourceType:
                self.validator.preprocessor.add_failure("[{0}] should not exist. Found in resource named {1}".format(self.requestedResourceType, terraformResource.name),
                                                        terraformResource.moduleName,
                                                        terraformResource.fileName,
                                                        self.validator.severity,
                                                        self.validator.isRuleOverridden,
                                                        self.validator.overrides,
                                                        terraformResource.name)

    def should_have_properties(self, properties_list):
        if type(properties_list) is not list:
            properties_list = [properties_list]

        if len(self.resource_list) > 0:
            for resource in self.resource_list:
                property_names = resource.config.keys()
                for required_property_name in properties_list:
                    if required_property_name not in property_names:
                        self.validator.preprocessor.add_failure("[{0}.{1}] should have property: '{2}'".format(resource.type, resource.name, required_property_name),
                                                                resource.moduleName,
                                                                resource.fileName,
                                                                self.validator.severity,
                                                                self.validator.isRuleOverridden,
                                                                self.validator.overrides,
                                                                resource.name)

    def should_not_have_properties(self, properties_list):
        if type(properties_list) is not list:
            properties_list = [properties_list]

        if len(self.resource_list) > 0:
            for resource in self.resource_list:
                property_names = resource.config.keys()
                for excluded_property_name in properties_list:
                    if excluded_property_name in property_names:
                        self.validator.preprocessor.add_failure("[{0}.{1}] should not have property: '{2}'".format(resource.type, resource.name, excluded_property_name),
                                                                resource.moduleName,
                                                                resource.fileName,
                                                                self.validator.severity,
                                                                self.validator.isRuleOverridden,
                                                                self.validator.overrides,
                                                                resource.name)

    def name_should_match_regex(self, regex):
        for resource in self.resource_list:
            if not self.validator.matches_regex_pattern(resource.name, regex):
                self.validator.preprocessor.add_failure("[{0}.{1}] name should match regex: '{2}'".format(resource.type, resource.name, regex),
                                                        resource.moduleName,
                                                        resource.fileName,
                                                        self.validator.severity,
                                                        self.validator.isRuleOverridden,
                                                        self.validator.overrides,
                                                        resource.name)


class Validator:

    # default severity is high
    severity = "high"
    preprocessor = None

    def __init__(self):
        self.raise_error_if_property_missing = False

    def resources(self, typ):
        resources = self.terraform.get('resource', {})

        return TerraformResourceList(self, typ, None, resources)

    def error_if_property_missing(self):
        self.raise_error_if_property_missing = True

    # generator that loops through all files to be scanned (stored internally in fileName; returns self (Validator) but sets self.fileName and self.terraform
    def get_terraform_files(self, isRuleOverridden):
        self.isRuleOverridden = isRuleOverridden
        for self.fileName, self.terraform in self.preprocessor.modulesDict.items():
            yield self

    def matches_regex_pattern(self, variable, regex):
        return not (self.get_regex_matches(regex, variable) is None)

    def get_regex_matches(self, regex, variable):
        if regex[-1:] != "$":
            regex = regex + "$"

        if regex[0] != "^":
            regex = "^" + regex

        variable = str(variable)
        if '\n' in variable:
            return re.match(regex, variable, re.DOTALL)
        return re.match(regex, variable)

    # this is only used by unit_test.py
    def setTerraform(self, terraform):
        self.terraform = terraform
        self.fileName = "none.tf"


class PreProcessor:

    TF = ".tf"
    UTF8 = "utf8"
    IS_MODULE = "__isModule__"
    PARENT = "__parent__"
    MODULE_NAME = "__ModuleName__"
    FILE_NAME = "__fileName__"
    LOCALS = "locals"
    VARIABLE = "variable"
    OUTPUT = "output"
    RESOURCE = "resource"
    VALUE = "value"
    MODULE = "module"
    SOURCE = "source"
    DEFAULT = "default"
    REGEX_COLON_BRACKET = re.compile('.*:\s*\[.*', re.DOTALL)           # any characters : whitespace [ any characters

    def __init__(self, jsonOutput):
        self.jsonOutput = jsonOutput
        self.variablesFromCommandLine = {}
        self.hclDict = {}
        self.modulesDict = {}
        self.fileNames = {}
        self.passNumber = 1
        self.dummyIndex = 0
        # on 1st pass replace var. with var$.
        # on 2nd pass replace var. and var$. with var!.
        self.braces = ["${", "@{", "!{"]
        self.vars = ["var.", "var@.", "var!."]
        self.locals = ["local.", "local@.", "local!."]
        self.modules = ["module.", "module@.", "module!."]
        self.terraform_workspaces = ["terraform.workspace", "terraform@.workspace", "terraform!.workspace"]
        self.datas = ["data.", "data@.", "data!."]
        self.variableFind = [self.braces[0], self.vars[0], self.locals[0], self.modules[0], self.terraform_workspaces[0], self.datas[0]]
        self.variableErrorReplacement = [self.braces[1], self.vars[1], self.locals[1], self.modules[1], self.terraform_workspaces[1], self.datas[1]]
        self.variableErrorReplacementPass2 = [self.braces[2], self.vars[2], self.locals[2], self.modules[2], self.terraform_workspaces[2], self.datas[2]]
        self.replacements = [self.variableFind, self.variableErrorReplacement, self.variableErrorReplacementPass2]
        self.replaceableVariablePrefixes = [self.braces[0], self.vars[0], self.locals[0], self.modules[0], self.terraform_workspaces[0],
                                            self.braces[1], self.vars[1], self.locals[1], self.modules[1], self.terraform_workspaces[1],
                                            self.braces[2], self.vars[2], self.locals[2], self.modules[2], self.terraform_workspaces[2]]
        
    def process(self, path, variablesJsonFilename=None):
        inputVars = {}
        if variablesJsonFilename is not None:
            for fileName in variablesJsonFilename:
                with open(fileName, "r", encoding="utf-8") as fp:
                    try:
                        variables_string = fp.read()
                        inputVarsDict = hcl.loads(variables_string)
                        inputVars = {**inputVars, **inputVarsDict}
                    except:
                        self.add_error_force(traceback.format_exc(), "---", fileName, "high")

        # prefix any input variable not containing '.' with 'var.'
        for var in inputVars:
            if "." not in var:
                newVar = "var." + var
                self.variablesFromCommandLine[newVar] = inputVars[var]
            else:
                self.variablesFromCommandLine[var] = inputVars[var]
                
        self.root = None
        self.readDir(path, self.hclDict)

        # all terraform files are now loaded into hclDict (indexed by subdirectory/fileName/terraform structure)
        # process hclDict and load every module into modulesDict
        self.getAllModules(self.hclDict, False)
        # make second pass so variables depending on a previous definition should now be defined
        logging.warning("------------------>>>starting pass 2...")
        self.passNumber = 2
        self.variableFind += self.variableErrorReplacement
        self.variableErrorReplacement = self.variableErrorReplacementPass2 + self.variableErrorReplacementPass2
        self.getAllModules(self.hclDict, False)

    def readDir(self, path, d):
        for directory, subdirectories, files in os.walk(path):
            if self.root is None:
                self.root = directory
                i = self.root.rfind(os.path.sep)
                if i != -1:
                    self.root = self.root[i+1:]
                # define root and mark this dictionary as a module since all directories in terraform are modules by default
                self.hclDict[self.root] = {}
                self.hclDict[self.root][self.IS_MODULE] = True
                self.hclDict[self.root][self.PARENT] = None
                self.hclDict[self.root][self.MODULE_NAME] = self.root
                d = self.hclDict[self.root]

            for file in files:
                if file[-3:].lower() == self.TF:
                    # terraform file (ends with .tf)
                    fileName = os.path.join(directory, file)
                    relativeFileName = fileName[len(path):]
                    with open(fileName, 'r', encoding='utf8') as fp:
                        try:
                            terraform_string = fp.read()
                            if len(terraform_string.strip()) > 0:
                                self.loadFileByDir(fileName, relativeFileName, d, d, terraform_string)
                                self.fileNames[fileName] = fileName
                        except:
                            self.add_error_force(traceback.format_exc(), "---", fileName, "high")

    # load file by directory, marking each directory as a module and setting parent directories
    def loadFileByDir(self, fileName, path, hclSubDirDict, parentDir, terraform_string):
        i = path.find("\\")
        if i == -1:
            # \ not found, try /
            i = path.find("/")
            if i == -1:
                # end of subdirectories; path is a terraform filename; load terraform file into dictionary
                hclSubDirDict[path] = hcl.loads(terraform_string)
                hclSubDirDict[path][self.FILE_NAME] = fileName
                # remove file name from end of path
                t = self.getPreviousLevel(fileName, os.path.sep)
                self.findModuleSources(hclSubDirDict[path], parentDir, t[0])
                return
        if i == 0:
            # found in first character, recursively try again skipping first character
            self.loadFileByDir(fileName, path[1:], hclSubDirDict, parentDir, terraform_string)
        else:
            # get subdirectory
            subdir = path[:i]
            if hclSubDirDict.get(subdir) is None:
                # subdirectory not defined in our dictionary yet so define it
                hclSubDirDict[subdir] = {}
                hclSubDir = hclSubDirDict[subdir]
                # mark this dictionary as a module since all directories in terraform are modules by default
                hclSubDir[self.IS_MODULE] = True
                hclSubDir[self.PARENT] = parentDir
                hclSubDir[self.MODULE_NAME] = subdir
            else:
                hclSubDir = hclSubDirDict[subdir]
            # recursively process next subdirectory
            i += 1
            self.loadFileByDir(fileName, path[i:], hclSubDirDict[subdir], hclSubDir, terraform_string)

    def findModuleSources(self, d, parentDir, currentFileName):
        for key in d:
            # only process module key
            if key == self.MODULE:
                modules = d[key]
                # process all modules
                for moduleName in modules:
                    module = modules[moduleName]
                    # find source parameter
                    for parameter in module:
                        if parameter == self.SOURCE:
                            sourcePath = module[parameter]
                            if not sourcePath.startswith("git::"):
                                self.createMissingFromSourcePath(sourcePath, parentDir, currentFileName)

    def createMissingFromSourcePath(self, sourcePath, d, currentFileName):
        # source is local
        while sourcePath != "":
            t = self.getNextLevel(sourcePath, "/")
            currentModule = t[0]
            sourcePath = t[1]
            if currentModule == "..":
                # move up a level
                t = self.getPreviousLevel(currentFileName, os.path.sep)
                currentFileName = t[0]
                if d.get(self.PARENT) is None:
                    # add parent
                    self.dummyIndex += 1
                    parent = {}
                    parent[self.PARENT] = None
                    parent[self.IS_MODULE] = True
                    parent[self.MODULE_NAME] = "dummy" + str(self.dummyIndex)
                    parent[d[self.MODULE_NAME]] = d
                    d[self.PARENT] = parent
                    self.hclDict = parent
                d = d[self.PARENT]
            elif currentModule == ".":
                # current directory; do nothing
                pass
            else:
                # move down to currentModule level
                currentFileName += os.path.sep + currentModule
                md = d.get(currentModule, False)
                if md is False:
                    # create new level
                    d[currentModule] = {}
                    md = d[currentModule]
                    md[self.PARENT] = d
                    md[self.MODULE_NAME] = currentModule
                    md[self.IS_MODULE] = True

                d = md
        # read directory
        self.readDir(currentFileName, d)
        return d

    # get all modules for given dictionary d; pass isModule as True if in a module block
    def getAllModules(self, d, isModule):
        for key in d:
            # ignore parent key
            if key != self.PARENT:
                value = d[key]
                if type(value) is dict:
                    if isModule or self.isModule(value):
                        moduleName = key
                        # load module, resolve variables in it and add it to modules dictionary
                        moduleDict = self.getModule(moduleName)
                        if self.isModule(value):
                            moduleDict[self.PARENT] = d[moduleName][self.PARENT]
                    # recursively get all modules in the nested dictionary in value
                    self.getAllModules(value, key == "module")

    # get given moduleName from modulesDict; find & load it if not there yet
    def getModule(self, moduleName, errorIfNotFound=True, dictToCopyFrom=None, tfDict=None):
        moduleDict = self.modulesDict.get(moduleName)
        if moduleDict is None or (moduleDict[self.VARIABLE] == {} and moduleDict[self.LOCALS] == {} and moduleDict[self.OUTPUT] == {}):
            # not there yet, find it and load it
            moduleDict = self.findModule(moduleName, self.hclDict, dictToCopyFrom, tfDict)
            if moduleDict is None:
                # couldn't find it, log it and create a dummy entry
                if errorIfNotFound:
                    self.logMsg("error", "Couldn't find module " + moduleName)
                moduleDict = self.createModuleEntry(moduleName)
        elif self.passNumber > 1:
            # module found on second pass, re-resolve variables
            self.findModule(moduleName, self.hclDict, dictToCopyFrom, tfDict)

        return moduleDict

    # find given moduleName in given dictionary d; load module attributes and resolve variables in module; last two parameters are all or nothing
    def findModule(self, moduleName, d, dictToCopyFrom=None, tfDict=None):
        # use dictToCopyFrom if provided
        if dictToCopyFrom is not None:
            sourcePath = self.getSourcePath(dictToCopyFrom)
            if sourcePath is not None and not sourcePath.startswith("git::"):
                dd = self.getModuleDictFromSourcePath(sourcePath, tfDict)
                if dd:
                    moduleDict = self.modulesDict.get(dd[self.MODULE_NAME])
                    if moduleDict is None:
                        moduleDict = self.createModuleEntry(dd[self.MODULE_NAME])
                        moduleDict[self.IS_MODULE] = True
                    self.loadModule(moduleName, dd, dictToCopyFrom)
                    if moduleName != dd[self.MODULE_NAME]:
                        self.loadModule(dd[self.MODULE_NAME], dd, dictToCopyFrom)
                    # source module found, replace variables and return it
                    m = self.loadModule(dd[self.MODULE_NAME], dd, dictToCopyFrom)
                    return m
            return self.loadModule(moduleName, {}, dictToCopyFrom)

        for key in d:
            # ignore parent key
            if key != self.PARENT:
                value = d[key]
                if key == moduleName:
                    # module found, replace variables and return it
                    return self.loadModule(moduleName, value, dictToCopyFrom)
                else:
                    if type(value) is dict:
                        if self.isModule(value):
                            # recursively find the module
                            m = self.findModule(moduleName, value)
                            if m is not None:
                                return m
        # not found
        return None

    def loadModule(self, moduleName, d, dictToCopyFrom):
        self.logMsgAlways("warning", ">>>loading module " + moduleName)

        moduleDict = self.modulesDict.get(moduleName)
        if moduleDict is None:
            # create empty module entry
            moduleDict = self.createModuleEntry(moduleName)
            moduleDict[self.IS_MODULE] = self.hasTerraform(d)

        if dictToCopyFrom is not None:
            mdv = moduleDict[self.VARIABLE]
            # add/replace the passed in variables to the module's variables
            for attr in dictToCopyFrom:
                if attr != self.SOURCE:
                    # only replace on pass #1 if resolved
                    if self.passNumber == 2 or self.isResolved(dictToCopyFrom[attr]):
                        mdv[attr] = dictToCopyFrom[attr]

        # load all attributes for this module
        self.loadModuleAttributes(moduleName, d, moduleDict, None)
        # resolve variables for this module
        self.resolveVariablesInModule(moduleName, moduleDict)
        return moduleDict

    def isResolved(self, var):
        if type(var) is str:
            return self.isStrResolved(var)
        elif type(var) is dict:
            for key in var:
                if not self.isResolved(var[key]):
                    return False
        elif type(var) is list:
            for value in var:
                if not self.isResolved(value):
                    return False
        else:
            return False
        return True

    def isStrResolved(self, var):
        for varErrorReplacement in self.variableErrorReplacement:
            if varErrorReplacement in var:
                return False
            
        return True

    def createModuleEntry(self, moduleName):
        self.modulesDict[moduleName] = {}
        moduleDict = self.modulesDict[moduleName]
        moduleDict[self.VARIABLE] = {}
        moduleDict[self.LOCALS] = {}
        moduleDict[self.OUTPUT] = {}
        moduleDict[self.RESOURCE] = {}
        moduleDict[self.IS_MODULE] = False
        return moduleDict

    def loadModuleAttributes(self, moduleName, d, moduleDict, tfDict):
        if self.isModule(d):
            if tfDict is None:
                tfDict = d
            else:
                # skip nested modules
                return

        for key in sorted(d):
            # ignore parent key
            if key != self.PARENT:
                value = d[key]
                if key == self.LOCALS:
                    # get values for all local variables
                    for local in value:
                        # only replace on first pass or not already fully resolved
                        if self.passNumber == 1 or self.containsVariable(moduleDict[self.LOCALS][local]):
                            moduleDict[self.LOCALS][local] = value[local]
                elif key == self.OUTPUT:
                    for output in value:
                        # only replace on first pass or not already fully resolved
                        if self.passNumber == 1 or self.containsVariable(moduleDict[self.OUTPUT][output]):
                            moduleDict[self.OUTPUT][output] = value[output][self.VALUE]
                elif key == self.RESOURCE:
                    for resourceType in value:
                        resourceNames = value[resourceType]
                        for resourceName in resourceNames:
                            config = resourceNames[resourceName]
                            res = moduleDict[self.RESOURCE].get(resourceName, None)
                            # only replace on first pass or not already fully resolved
                            if self.passNumber == 1 or (res != None and self.containsVariable(res.config)):
                                moduleDict[self.RESOURCE][resourceName] = TerraformResource(resourceType, resourceName, config, d[self.FILE_NAME], moduleName)
                elif key == self.VARIABLE:
                    '''
                    value could be a string as in below case
                        condition {
                            test = "ArnEquals"
                            variable = "aws:SourceArn"
                            values = ["${var.services_entry_arn}"]
                        }
                    '''
                    if type(value) is dict:
                        # initialize any default values for variables
                        for variable in value:
                            if value[variable].get(self.DEFAULT) is not None:
                                moduleDict[self.VARIABLE][variable] = value[variable][self.DEFAULT]
                elif key == self.MODULE:
                    # loop through all modules
                    for mn in value:
                        # resolve parameter variables first
                        for parameter in value[mn]:
                            if parameter != self.SOURCE:
                                replacementValue = self.resolveVariableByType(value[mn][parameter], moduleName)
                                if replacementValue != value[mn][parameter]:
                                    self.logMsgAlways("warning", "replaced module " + mn + " parameter " + parameter + " value " + str(value[mn][parameter]) + " with " + str(replacementValue))
                                    value[mn][parameter] = replacementValue
                        # get defined module; load it if not already there
                        md = self.getModule(mn, False, value[mn], tfDict)
                        # copy all outputs from source module (md) to containing module variable
                        for output in md[self.OUTPUT]:
                            # only copy if not already there
                            if moduleDict[self.VARIABLE].get(output) is None:
                                moduleDict[self.VARIABLE][output] = md[self.OUTPUT][output]
                else:
                    if type(value) is dict:
                        # don't load any other nested modules
                        if not self.isModule(value):
                            self.loadModuleAttributes(moduleName, value, moduleDict, tfDict)

    def getSourcePath(self, parameterDict):
        for parameter in parameterDict:
            if parameter == self.SOURCE:
                return parameterDict[parameter]
        return None

    def getModuleDictFromSourcePath(self, sourcePath, d):
        # source is local
        while sourcePath != "":
            t = self.getNextLevel(sourcePath, "/")
            currentModule = t[0]
            sourcePath = t[1]
            if currentModule == "..":
                # move up a level
                d = d[self.PARENT]
            elif currentModule == ".":
                # current directory; do nothing
                pass
            else:
                # move down to currentModule level
                d = d.get(currentModule, False)
                if d is False:
                    return False
        return d

    # resolve variables (anything surrounded by ${}) in given moduleDict
    def resolveVariablesInModule(self, moduleName, moduleDict):
        self.shouldLogErrors = False
        # resolve variables
        for key in moduleDict[self.VARIABLE]:
            value = moduleDict[self.VARIABLE][key]
            replacementValue = self.resolveVariableByType(value, moduleName)
            moduleDict[self.VARIABLE][key] = replacementValue
            if replacementValue != value:
                self.logMsgAlways("warning", "replaced variable " + key + " value " + str(value) + " with " + str(replacementValue))
        # resolve locals
        for key in moduleDict[self.LOCALS]:
            value = moduleDict[self.LOCALS][key]
            replacementValue = self.resolveVariableByType(value, moduleName)
            moduleDict[self.LOCALS][key] = replacementValue
            if replacementValue != value:
                self.logMsgAlways("warning", "replaced local variable " + key + " value " + str(value) + " with " + str(replacementValue))
        # resolve outputs
        for key in moduleDict[self.OUTPUT]:
            value = moduleDict[self.OUTPUT][key]
            replacementValue = self.resolveVariableByType(value, moduleName)
            moduleDict[self.OUTPUT][key] = replacementValue
            if replacementValue != value:
                self.logMsgAlways("warning", "replaced output variable " + key + " value " + str(value) + " with " + str(replacementValue))
        # resolve resources
        self.shouldLogErrors = True
        for key in moduleDict[self.RESOURCE]:
            value = moduleDict[self.RESOURCE][key].config
            replacementValue = self.resolveVariableByType(value, moduleName)
            moduleDict[self.RESOURCE][key].config = replacementValue
            if replacementValue != value:
                self.logMsgAlways("warning", "replaced resource variable " + key + " value " + str(value) + " with " + str(replacementValue))

    def resolveVariableByType(self, value, moduleName):
        if type(value) is str:
            return self.resolveVariableLine(value, moduleName)
        elif type(value) is dict:
            return self.resolveDictVariable(value, moduleName)
        elif type(value) is list:
            return self.resolveListVariable(value, moduleName)
        elif type(value) is tuple:
            return self.resolveTupleVariable(value, moduleName)
        else:
            return value

    def resolveDictVariable(self, value, moduleName):
        returnValue = {}
        for key in value:
            returnValue[key] = self.resolveVariableByType(value[key], moduleName)
        return returnValue

    def resolveListVariable(self, value, moduleName):
        if len(value) == 0:
            return value;
        index = 0
        for v in value:
            value[index] = self.resolveVariableByType(v, moduleName)
            index += 1
        if value[0] in ("join", "merge", "concat", "coalesce", "element", "coalescelist"):
            # supported function
            return self.handleFunction(value)
        return value

    def resolveTupleVariable(self, value, moduleName):
        returnValue = tuple(self.resolveVariableByType(v, moduleName) for v in value)
        if len(returnValue) == 3:
            floatValue0 = self.getFloatValue(returnValue[0])
            floatValue2 = self.getFloatValue(returnValue[2])
            if type(floatValue0) == float and type(returnValue[1]) == str and type(floatValue2) == float:
                if returnValue[1] == "+":
                    return floatValue0 + floatValue2
                elif returnValue[1] == "-":
                    return floatValue0 - floatValue2
                elif returnValue[1] == "*":
                    return floatValue0 * floatValue2
                elif returnValue[1] == "/":
                    return floatValue0 / floatValue2
        return returnValue
            
    def getFloatValue(self, value):
        try:
            return float(value)
        except:
            return value;

    def handleFunction(self, value):
        # check if all variables have been resolved
        if self.containsVariable(value, True):
            # not fully resolved yet; return what we have so far
            return value
        return self.processFunction(value)
        
    def processFunction(self, value):
        it = iter(value)
        function = next(it, None)
        if function == "join":
            delimiter = next(it, None)
            t = next(it, None)
            return delimiter.join(v for v in t)
        elif function == "merge":
            d = {}
            for v in it:
                if type(v) is dict:
                    for key in v:
                        d[key] = v[key]
                else:
                    d[v] = v
            return d;
        elif function == "concat":
            d = []
            for v in it:
                if type(v) is list:
                    for entry in v:
                        d.append(entry)
                else:
                    d.append(v)
            return d;
        elif function == "element":
            lst = next(it, None)
            index = next(it, None)
            if '*' in lst:
                return value
            return lst[index]
        else:
            # coalesce/coalescelist
            d = {}
            if value[len(value)-1] == "...":
                # there is a single list that needs to be processed
                for v in value[1]:
                    if v:
                        if type(v) is tuple:
                            return self.processFunction(v)
                        else:
                            return v
            else:
                for v in it:
                    if v:
                        if type(v) is tuple:
                            return self.processFunction(v)
                        else:
                            return v
            # no non-empty entries; undefined what to do so return None
            return None;
        

    # returns True if given dictionary d contains a key of __isModule__
    def isModule(self, d):
        for key in d:
            if key == self.IS_MODULE:
                return d[key]
        return False

    # returns True if given dictionary d contains at least one terraform file
    def hasTerraform(self, d):
        for key in d:
            if key.lower().endswith(self.TF):
                return True
        return False

    # resolve entire variable
    def resolveVariableLine(self, value, moduleName):
        if not self.containsVariable(value):
            return value
        # a variable needs to be replaced
        t = self.findVariable(value, True)
        var = t[0]
        b = t[1]
        e = t[2]
        if var.startswith("["):
            var = var[1:len(var)-1]
        var = var.strip()
        rv = self.resolveVariable(var, moduleName)
        if b == 0 and (e == len(value) or e == -1):
            # full replacement; don't merge since a string may not have been returned
            newValue = rv[0]
        else:
            newValue = value[:b] + str(rv[0]) + value[e:]
        # recursively resolve the variables since there may be more than one variable in this value
        return self.resolveVariableByType(newValue, moduleName)

    # resolve innermost variable
    def resolveVariable(self, value, moduleName, dictToCopyFrom=None, tfDict=None):
        # find variable (possibly in brackets)
        isOldTFvarStyle=False
        v, b, e, insideBrackets, foundDelineator, foundDelineatorErrRepl = self.findVariable(value, False)
        if len(v) > 1 and v[1] == "{":
            isOldTFvarStyle = True
        if not insideBrackets and isOldTFvarStyle:
            # inside ${}; remove them
            var = value[2:e-1]
        else:
            var = v
        # update moduleName in case we switch modules and need to recurse more
        replacementValue, moduleName, isHandledType = self.getReplacementValue(var, moduleName, isOldTFvarStyle, dictToCopyFrom, tfDict)
        if replacementValue == var:
            # couldn't find a replacement; change to our notation to mark it
            if isHandledType:
                self.logMsg("error", "Couldn't find a replacement for: " + self.getOrigVar(var) + " in " + moduleName)
            else:
                self.logMsg("debug", "Couldn't find a replacement for: " + self.getOrigVar(var) + " in " + moduleName)
            if not isOldTFvarStyle:
                # strip off replaceable variable
                var = var[len(foundDelineator):]
            replacementValue = value[:b] + foundDelineatorErrRepl + var
            if insideBrackets:
                replacementValue += "]"
            if not isOldTFvarStyle and len(value) > 1 and value[1] == "{":
                replacementValue += "}"
            if isOldTFvarStyle and e > 0:
                # remove closing brace
                replacementValue += value[e-1:]
            return (replacementValue, not insideBrackets)

        if type(replacementValue) is str:
            if insideBrackets:
                self.logMsgAlways("info", "  replacing [" + var + "] with " + replacementValue)
                # resolve the variable again since the replacement may also contain variables
                return (value[:b] + self.resolveVariableLine(replacementValue, moduleName) + value[e:], not insideBrackets)
            else:
                if v == replacementValue:
                    # this prevents a loop
                    replacementValue = replacementValue.replace(foundDelineator, foundDelineatorErrRepl, 1)
                    self.logMsg("debug", "Couldn't find a replacement for: " + self.getOrigVar(var) + " (would have looped) in " + moduleName)
                else:
                    self.logMsgAlways("info", "  replacing ${" + var + "} with " + replacementValue)
                    # resolve the variable again since the replacement may also contain variables
                    return (self.resolveVariableLine(replacementValue, moduleName), not insideBrackets)
        else:
            if isOldTFvarStyle:
                self.logMsgAlways("info", "  replacing " + foundDelineator + var + "} with " + str(replacementValue))
            else:
                self.logMsgAlways("info", "  replacing " + var + " with " + str(replacementValue))

        return (replacementValue, not insideBrackets)

    def getOrigVar(self, var):
        if var.startswith(self.vars[1]) or var.startswith(self.vars[2]):
            return self.vars[0] + var[len(self.vars[1]):]
        elif var.startswith(self.locals[1]) or var.startswith(self.locals[2]):
            return self.locals[0] + var[len(self.locals[1]):]
        elif var.startswith(self.modules[1]) or var.startswith(self.modules[2]):
            return self.modules[0] + var[len(self.modules[1]):]
        else:
            return var

    # check if given value contains a variable anywhere
    def containsVariable(self, value, isAnyVar=False):
        if type(value) is str:
            t = self.findAnyVariableDelineatorsForVars(value, False, isAnyVar)
            if t[0] == -1:
                return False
            return True
        elif type(value) is dict:
            return self.containsVariableDict(value, isAnyVar)
        elif type(value) is list:
            return self.containsVariableList(value, isAnyVar)
        else:
            # not a variable
            return False

    def containsVariableDict(self, value, isAnyVar):
        for key in value:
            if self.containsVariable(value[key], isAnyVar):
                return True
        # no variables found
        return False

    def containsVariableList(self, value, isAnyVar):
        for v in value:
            if self.containsVariable(v, isAnyVar):
                return True
        # no variables found
        return False

    # find deepest nested variable in given value
    def findVariable(self, value, isNested, previouslyFoundVar=None):
        # pass 1: if unreplaceable, change $ to @
        # pass 2: if unreplaceable, change both $ & @ to !
        if type(value) is str:
            isVar = False
            val = value
            if previouslyFoundVar:
                insideBrackets = previouslyFoundVar[3]
            else:
                insideBrackets = False
            if isNested and type(previouslyFoundVar) is str and "{" in previouslyFoundVar[0]:
                # if this is a nested call and the outer call found ${, only look for the brace now
                braceOnly = True
            else:
                braceOnly = False
 
            b, e, foundDelineator, foundDelineatorErrRepl = self.findVariableDelineatorsForVars(val, braceOnly, self.variableFind, self.variableErrorReplacement)
            if b == -1:
                return None                
            if b > 0:
                partial = value[:b]
                if partial[len(partial)-1] == "[":
                    # open bracket found before the variable
                    insideBrackets = True
                    partial = value[b:e]
                    if partial[len(partial)-1] == "]":
                        # close bracket found after the variable, remove from variable end
                        e -= 1
                
            isVar = True
            if "{" in foundDelineator and e == -1:
                # problem
                self.add_error("Matching close brace not found: " + value, "---", "---", "high")
                return None
            
            foundVar = (value[b:e], b, e, insideBrackets, foundDelineator, foundDelineatorErrRepl)
            
            newSearchValue = foundVar[0]
            bOffset = 0
            if isVar:
                # remove delineator(s)
                if newSearchValue.endswith("}"):
                    newSearchValue = newSearchValue[2:len(newSearchValue)-1]
                    bOffset = 2
                else:
                    newSearchValue = newSearchValue[len(foundDelineator):]
                    bOffset = len(foundDelineator)
            else:
                newSearchValue = newSearchValue[1:len(newSearchValue)-1]
                if not insideBrackets:
                    # adjust beginning & ending since nested inside previouslyFoundVar
                    fv_length = len(previouslyFoundVar[4])
                    b += fv_length
                    e += fv_length
                    bOffset = 0
                else:
                    bOffset = 1
                foundVar = (newSearchValue, b, e, insideBrackets, foundDelineator, foundDelineatorErrRepl)
            if newSearchValue not in self.terraform_workspaces[0]:
                # recursively find variable
                fv = self.findVariable(newSearchValue, True, foundVar)
                if fv is None:
                    # no variable found
                    return foundVar
                if foundVar[0].endswith("}") and fv[1] == 0 and fv[2] == len(newSearchValue):
                    # return originally found variable which is the old style
                    return foundVar
                if insideBrackets:
                    # use beginning & ending from previous
                    fv = (fv[0], b, e, fv[3], fv[4], fv[5])
                else:
                    fv = (fv[0], fv[1]+b+bOffset, fv[2]+b+bOffset, fv[3], fv[4], fv[5])
                return fv
            else:
                return foundVar
        return previouslyFoundVar

    def findAnyVariableDelineatorsForVars(self, value, braceOnly, isAnyVar):
        if isAnyVar:
            for variableFind in self.replaceableVariablePrefixes:
                t = self.findVariableDelineatorsForVars(value, braceOnly, [variableFind], [variableFind])
                if t[0] != -1:
                    return t
            return -1, 0, None, None
        return self.findVariableDelineatorsForVars(value, braceOnly, self.variableFind, self.variableErrorReplacement)
        
    def findVariableDelineatorsForVars(self, value, braceOnly, variableFind, variableErrorReplacement):
        if braceOnly and value not in self.terraform_workspaces:
            b, e = self.findVariableDelineators(value, variableFind[0], "}", variableErrorReplacement[0])
            if b > -1:
                return b, e, variableFind[0], variableErrorReplacement[0]
        else:
            prevB = -1
            for varPrefix, varErrorReplacement in zip(variableFind, variableErrorReplacement):
                if varPrefix[1] == "{":
                    closeVar = "}"
                else:
                    closeVar = None
                b, e = self.findVariableDelineators(value, varPrefix, closeVar, varErrorReplacement)
                if b > -1:
                    if closeVar != None or b == 0 or (value[b-1] != "{"):
                        if b > prevB:
                            prevB = b;
                            prevE = e
                            prevVarPrefix = varPrefix
                            prevVarErrorReplacement = varErrorReplacement
            if prevB != -1:
                return prevB, prevE, prevVarPrefix, prevVarErrorReplacement
        return -1, 0, None, None

    def findVariableDelineators(self, value, openVar, closeVar, varErrorReplacement=None):
        '''
        This is valid:  name-prefix = "sf-${module.common.account_name}-${local.pcas_vpc_type}-${local.env}-${module.common.region}"
        This is not:    name-prefix = "sf-module.common.account_name-local.pcas_vpc_type-local.env-module.common.region"
        i.e. if combining variables into one variable, must use old ${} variable style
        '''
        b = value.rfind(openVar)
        if b == -1:
            return -1, 0
        if openVar == "[":
            # check if preceeded by :
            matchObject = self.REGEX_COLON_BRACKET.search(value)
            if matchObject != None:
                return -1, 0
        if closeVar == None:
            if openVar in self.terraform_workspaces:
                return b, b + len(openVar)
            # search for default "closeVars"
            if value.find("[", b) == -1:
                defaultCloseVars = [",", "'", ")", "}", "]"]
            else:
                # don't include ] if [ is in value
                defaultCloseVars = [",", "'", ")", "}"]
            prevE = 99999
            for closeVar in defaultCloseVars:
                e = value.find(closeVar, b)
                if e != -1 and e < prevE:
                    prevE = e
            if prevE != 99999:
                return b, prevE
            # no close value found, use length of value
            return b, len(value)
        v = value[b+1:]
        nested = 0
        for index, char in enumerate(v):
            if char == closeVar:
                if nested == 0:
                    return b, b+index+2
                # just closed a nested variable
                nested -= 1
            if v[index:index+len(openVar)] == openVar or (varErrorReplacement != None and len(v) >= index+len(varErrorReplacement) and v[index:index+len(varErrorReplacement)] == varErrorReplacement):
                # start of a nested variable
                nested += 1
        # error: matching closeVar not found
        return 0, -1

    # find replacement value for given var in given moduleName
    def getReplacementValue(self, var, moduleName, isOldTFvarStyle, dictToCopyFrom=None, tfDict=None):
        replacementValue = None
        if (var.startswith('"') and var.endswith('"')) or (var.startswith("'") and var.endswith("'")):
            return var[1:len(var)-1], moduleName, True
        subscript = None
        b = var.find("[")
        if b != -1:
            v = var[:b]
            e = var.find("]", b)
            if e != -1:
                subscript = var[b+1:e]
                v += var[e+1:]
                if subscript[0] == '"' or subscript[0] == "'":
                    # remove quotes
                    subscript = subscript[1:len(subscript)-1]
        else:
            v = var

        isHandledType = False
        notHandled = ["?", "==", "!=", ">", "<", ">=", "<=", "&&", "||", "!", "+", "-", "*", "/", "%"]
        moduleDict = self.modulesDict[moduleName]
        if isOldTFvarStyle:
            varsTuple = self.vars[0]
            localsTuple = self.locals[0]
            modulesTuple = self.modules[0]
            terraform_workspacesTuple = self.terraform_workspaces[0]
        else:
            if self.passNumber == 1:
                varsTuple = self.vars[0]
                localsTuple = self.locals[0]
                modulesTuple = self.modules[0]
                terraform_workspacesTuple = self.terraform_workspaces[0]
            else:
                varsTuple = (self.vars[0], self.vars[1])
                localsTuple = (self.locals[0], self.locals[1])
                modulesTuple = (self.modules[0], self.modules[1])
                terraform_workspacesTuple = (self.terraform_workspaces[0], self.terraform_workspaces[1])
        if v.startswith(varsTuple):
            # conditional statements, boolean statements, and math are not currently handled
            if not any(x in v for x in notHandled):
                isHandledType = True
            v = v[self.getPrefixLength(v, varsTuple):]
            index = v.find('.')
            if index > -1:
                subscript = v[index+1:]
                v = v[:index]
            replacementValue = moduleDict[self.VARIABLE].get(v)
        elif v.startswith(localsTuple):
            if not any(x in v for x in notHandled):
                isHandledType = True
            v = v[self.getPrefixLength(v, localsTuple):]
            replacementValue = moduleDict[self.LOCALS].get(v)
        elif v.startswith(modulesTuple):
            if not any(x in v for x in notHandled):
                isHandledType = True
            # variable is in a different module
            modulePrefixLength = self.getPrefixLength(v, modulesTuple)
            e = v.find(".", modulePrefixLength)
            if e == -1:
                self.add_error("Error Resolving module variable: " + var + "  expected ending '.' not found", moduleName, "---", "high")
            else:
                moduleName = v[modulePrefixLength:e]
                md = self.getModule(moduleName, True, dictToCopyFrom, tfDict)
                moduleOutputDict = md[self.OUTPUT]
                e += 1
                remainingVar = v[e:]
                while remainingVar != "":
                    t = self.getNextLevel(remainingVar, ".")
                    moduleOutputDict = moduleOutputDict.get(t[0])
                    if moduleOutputDict is None:
                        self.logMsg("error", "Error resolving variable: " + var + "  variable not found in module (no module source available?) in " + moduleName)
                        return var, moduleName, True
                    remainingVar = t[1]
                if type(moduleOutputDict) is dict:
                    replacementValue = moduleOutputDict.get(self.VALUE, moduleOutputDict)
                else:
                    replacementValue = moduleOutputDict
        elif v.startswith(terraform_workspacesTuple):
            isHandledType = True

        if type(replacementValue) is dict and subscript != None:
            replacementValue = replacementValue.get(subscript, subscript)

        if replacementValue is None:
            replacementValue = self.variablesFromCommandLine.get(var, var)

        return replacementValue, moduleName, isHandledType

    def getPrefixLength(self, var, varTuple):
        if type(varTuple) is str:
            return len(varTuple)
        for t in varTuple:
            if var.startswith(t):
                return len(t)

    def getPreviousLevel(self, var, separator):
        b = var.rfind(separator)
        if b == -1:
            b = len(var)
        return (var[:b], var[b+1:])

    def getNextLevel(self, var, separator):
        b = var.find(separator)
        if b == -1:
            b = len(var)
        return (var[:b], var[b+1:])

    # add given failure in given fileName
    def add_failure(self, failure, moduleName, fileName, severity, isRuleOverridden, overrides, resourceName):
        waived = ""
        waiver = self.overridden(isRuleOverridden, overrides, resourceName, severity)
        if waiver is not None:
            if severity == "high":
                waived = "**waived by " + waiver + "**"
            else:
                waived = "**waived**"
        self.jsonOutput["failures"].append( self.getFailureMsg(severity, waived, failure, moduleName, fileName) )

    def overridden(self, isRuleOverridden, overrides, resourceName, severity):
        if isRuleOverridden:
            for override in overrides:
                if override[1] == resourceName:
                    if severity == "high":
                        if len(override) == 3:
                            return override[2]
                        print("***Invalid override: " + ":".join(override))
                        print("high severity rules must include RR or RAR")
                        print("Needs to be in the following format:  rule_name:resource_name:RR-xxx or rule_name:resource_name:RAR-xxx where xxx is 1-10 digits")
                        sys.exit(99)
                    return ""
        return None

    # add given error in given fileName
    def add_error(self, error, moduleName, fileName, severity):
        if self.passNumber == 2 and self.shouldLogErrors:
            self.add_error_force(error, moduleName, fileName, severity)

    def add_error_force(self, error, moduleName, fileName, severity):
        self.jsonOutput["errors"].append( self.getFailureMsg(severity, "", error, moduleName, fileName) )

    def getFailureMsg(self, severity, waived, msg, moduleName, fileName):
        message = {}
        message["severity"] = severity
        message["waived"] = waived
        message["message"] = msg
        message["moduleName"] = moduleName
        message["fileName"] = fileName
        return message

    def logMsg(self, typ, msg):
        if self.passNumber == 2:
            self.logMsgAlways(typ, msg)

    def logMsgAlways(self, typ, msg):
        if typ == "error":
            logging.error(msg)
        elif typ == "warning":
            logging.warning(msg)
        elif typ == "info":
            logging.info(msg)
        elif typ == "debug":
            logging.debug(msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

readme = '''=========
Terrascan
=========

.. image:: https://img.shields.io/pypi/v/terrascan.svg
        :target: https://pypi.python.org/pypi/terrascan
        :alt: pypi

.. image:: https://img.shields.io/travis/cesar-rodriguez/terrascan.svg
        :target: https://travis-ci.org/cesar-rodriguez/terrascan
        :alt: build

.. image:: https://readthedocs.org/projects/terrascan/badge/?version=latest
        :target: https://terrascan.readthedocs.io/en/latest/?badge=latest
        :alt: Documentation Status

.. image:: https://pyup.io/repos/github/cesar-rodriguez/terrascan/shield.svg
     :target: https://pyup.io/repos/github/cesar-rodriguez/terrascan/
     :alt: Updates


A collection of security and best practice tests for static code analysis of terraform_ templates using terraform_validate_.

.. _terraform: https://www.terraform.io
.. _terraform_validate: https://github.com/elmundio87/terraform_validate

* GitHub Repo: https://github.com/cesar-rodriguez/terrascan
* Documentation: https://terrascan.readthedocs.io.
* Free software: GNU General Public License v3
'''

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'pyhcl==0.4.0',
]

setup(
    name='terrascan',
    version='0.2.0',
    description="Best practices tests for terraform",
    long_description=readme,
    author="Cesar Rodriguez",
    author_email='therasec@gmail.com',
    url='https://github.com/cesar-rodriguez/terrascan',
    download_url='https://github.com/cesar-rodriguez/terrascan' +
    '/archive/v0.2.0.tar.gz',
    packages=find_packages(where='.'),
    entry_points={
        'console_scripts': [
            'terrascan = terrascan.terrascan:main'
        ]
    },
    include_package_data=True,
    license="GNU General Public License v3",
    zip_safe=False,
    keywords='terrascan',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    test_suite='tests',
    tests_require=requirements,
    setup_requires=requirements,
    install_requires=requirements,
)
